/* Created by Anjuta version 1.1.98 */
/*	This file will not be overwritten */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <allegro.h>

#include "crc32.h"
#include "audio.h"
#include "vmachine.h"
#include "vdc.h"
#include "cpu.h"
#include "debug.h"
#include "keyboard.h"
#include "voice.h"

#define Versione			"0.1"
#define pScanLines		"scl"
#define NoSound			"nsnd"
#define NoVoice				"nvoice"
#define SndVol				"sndvol"
#define VcVol					"vcvol"
#define pFilter 				"lpf"
#define pDebug				"debug"
#define pSpeed				"speed"
#define NoLimit				"nolim"
#define pBios					"bios"
#define pEuro					"euro"
#define pExRom				"erom"
#define pStick					"stick"
#define pFps					"fps"
#define pFullScr				"fullscr"

#define MAXC 				1024

static char Bios[MAXC],ScShot[MAXC];


static long FileSize(FILE *S);
static void Load_Bios(const char *BiosName);
static void Load_Cart(char *File);
int Parse_Option(char *Attr,char *Val);
void Read_Default_Config(void);

int main(int argc, char *argv[])
{
	int i;
	static char file[MAXC], attr[MAXC], val[MAXC], *p;
	
	printf("loem ver %s .:Linux Odyssy2 EMulator:.\n",Versione);
	printf("Based on O2EM by Andre de la Rocha\n");
	printf("loem is written by Calogiuri Enzo Antonio\n");
	printf("please mail me at enzo76@freemail.it\n");
	printf("-----------------------------------------------\n");
	
	printf("Emulator Function Keys\n");
	printf("	F1  - Info Screen\n");
	printf("	F4  - Debug Mode\n");
	printf("	F5  - Reset Emulator\n");
	printf("	F8  - Save a .pcx screenshot\n");
	printf("	F11 - Quit Emulator\n");
	printf("\n");
	
	if (argc < 2)
	{
		printf("Wrong number of parameters...\n");
		printf("use: loem <file> [Options]\n");
		printf("  <file>  	= file to load with ext\n");
		printf("  -%s    	= Enable scanlines\n",pScanLines);
		printf("  -%s    	= Start in fullscreen mode\n",pFullScr);
		printf("  -%s   	= Disable sound\n",NoSound);
		printf("  -%s 	= Disable voice emulation\n",NoVoice);
		printf("  -%s 	= Set sound volume[0..100]\n",SndVol);
		printf("  -%s  	= Set voice volume[0..100]\n",VcVol);
		printf("  -%s    	= Enable low-pass audio filter\n",pFilter);
		printf("  -%s  	= Show fps\n",pFps);
		printf("  -%s  	= Start loem in debug mode\n",pDebug);
		printf("  -%s=n  	= Relative speed (100 = original)\n",pSpeed);
		printf("  -%s  	= Turn off speed limit\n",NoLimit);
		printf("  -%s=file  	= Set bios name (default 02rom.bin)\n",pBios);		
		printf("  -%s  	= Use European timing mode (50Hz)\n",pEuro);
		printf("  -%s  	= Use special 3Kb program/1Kb data ROM mode\n",pExRom);
		printf("  -%s<n>=mode/keys  	= Define stick n mode/keys (N=1-2)\n",pStick);		
		
		fflush(stdout);
		exit(EXIT_SUCCESS);
	}
	
	App_Data.Debug=0;
	App_Data.Stick[0]=App_Data.Stick[1]=1;	
	
	Set_DefJoyKeys(0,0);
	Set_DefJoyKeys(1,1);
	
	App_Data.Bank=0;
	App_Data.Limit=1;
	App_Data.Sound_En=1;
	App_Data.Speed=100;
	App_Data.WSize=2;
	App_Data.FullScreen=0;
	App_Data.ScanLines=0;
	App_Data.Voice=1;
	App_Data.Window_Title="loem ver. " Versione;
	App_Data.SVolume=100;
	App_Data.VVolume=100;
	App_Data.Filter=0;
	App_Data.ExRom=0;
	App_Data.Crc=0;
	App_Data.ScShot=ScShot;
	App_Data.Euro=0;
	App_Data.OpenB=0;
	Show_Fps=0;
	
	strcpy(file,"");
	strcpy(Bios,"");
	strcpy(ScShot,"");
	
	Read_Default_Config();
	
	for(i=1; i < argc; i++)
	{
		if (argv[i][0] != '-')
		{
			strncpy(file,argv[i],MAXC-1);
			file[MAXC-1]=0;
		}
		else
		{
			p=strtok(argv[i],"=");
			
			if (p)
			{
				strncpy(attr,p+1,MAXC-1);
				attr[MAXC-1]=0;
			}
			else
				strcpy(attr,"");
			
			p=strtok(NULL,"=");
			
			if (p)
			{
				strncpy(val,p,MAXC-1);
				val[MAXC-1]=0;
			}
			else
				strcpy(val,"");
			
			strlwr(attr);
			
			if (!Parse_Option(attr, val))
				exit(EXIT_FAILURE);
		}
	}
	
	if (strlen(file) == 0)
	{
		fprintf(stderr,"Error: file name missing\n");
		exit(EXIT_FAILURE);
	}
	
	printf("Starting emulation ...\n");
	
	allegro_init();
	install_timer();
	Init_Audio();
	
	printf("Using Allegro Library %s\n",allegro_id);
	
	Load_Bios(Bios);
	Load_Cart(file);
	
	if (App_Data.Voice)
		Load_Voice_Samples();
	
	Init_Display();
	Init_Cpu();
	Init_System();
	
	if (App_Data.Debug)
		Key_Debug=1;
	
	Run();
	
	exit(EXIT_SUCCESS);
}

END_OF_MAIN();

int Parse_Option(char *Attr,char *Val)
{
	int t;
	
	if (!strcmp(Attr,NoLimit))
		App_Data.Limit=!(Val[0]!='0');
	else
		if (!strcmp(Attr,NoSound))
			App_Data.Sound_En=!(Val[0]!='0');
		else
			if (!strcmp(Attr,NoVoice))
				App_Data.Voice=!(Val[0]!='0');
			else
				if (!strcmp(Attr,pFilter))
					App_Data.Filter=(Val[0]!='0');
				else
					if (!strcmp(Attr,pDebug))
						App_Data.Debug=(Val[0]!='0');
					else
						if ((!strcmp(Attr,"stick1")) || (!strcmp(Attr,"stick2")))
						{
							int sn;
							
							sn = (!strcmp(Attr,"stick1"))? 0 : 1;
							
							if (strlen(Val)<2)
							{
								t=-1;
								sscanf(Val,"%d",&t);
								
								if ((t >= 0) && (t <= 3))
								{
									if ((t == 1) || (t == 2))
									{
										App_Data.Stick[sn]=1;
										
										Set_DefJoyKeys(sn,t - 1);
									}
									else
									{
										App_Data.Stick[sn]=(t == 0) ? 0 : 2;
										
										Set_JoyKeys(sn,0,0,0,0,0);
									}
								}
								else
								{
									fprintf(stderr,"Invalid value for option %s\n",Attr);
									return 0;
								}
							}
							else
							{
								char *p,*s;
								int i,k,code,nk,codes[5];
								
								strupr(Val);
								nk=0;
								p=strtok(Val,",");
								
								while (p)
								{
									i=code=0;
									k=KeybTab[i].KeybCode;
									s=KeybTab[i].KeybName;
									
									while (k && (code == 0))
									{
										if (strcmp(s,p)==0)
											code=k;
										
										i++;
										
										k=KeybTab[i].KeybCode;
										s=KeybTab[i].KeybName;
									}
									
									if (!code)
									{
										fprintf(stderr,"Invalid value for option %s : key %s unknown\n",Attr,p);
										return 0;
									}
									
									codes[nk]=code;
									p=strtok(NULL,",");
									nk++;
									
									if (nk > 5)
									{
										fprintf(stderr,"Invalid value for option %s : invalid number of keys\n",Attr);
										return 0;
									}
								}
								
								if (nk != 5)
								{
									fprintf(stderr,"Invalid value for option %s : invalid number of keys\n",Attr);
									return 0;
								}
								
								App_Data.Stick[sn]=1;
								
								Set_JoyKeys(sn,codes[0],codes[1],codes[2],codes[3],codes[4]);
							}
						}
						else
							if (!strcmp(Attr,pSpeed))
							{
								t=-1;
								sscanf(Val,"%d",&t);
								
								if ((t > 0) && (t <= 10000))
									App_Data.Speed=t;
								else
								{
									fprintf(stderr,"Invalid value for option %s\n",Attr);
									return 0;
								}
							}
							else
								if (!strcmp(Attr,SndVol))
								{
									t=-1;
									sscanf(Val,"%d",&t);
									
									if ((t >= 0) && (t <= 100))
										App_Data.SVolume=t;
									else
									{
										fprintf(stderr,"Invalid value for option %s\n",Attr);
										return 0;
									}
									
									if (t==0)
										App_Data.Sound_En=0;
								}
								else
									if (!strcmp(Attr,VcVol))
									{
										t=-1;
										sscanf(Val,"%d",&t);
										
										if ((t >= 0) && (t <= 100))
											App_Data.VVolume=t;
										else
										{
											fprintf(stderr,"Invalid value for option %s\n",Attr);
											return 0;
										}
										
										if (t==0)
											App_Data.Voice=0;
									}
									else
										if (!strcmp(Attr,pScanLines))
											App_Data.ScanLines=(Val[0]!='0');
										else
											if (!strcmp(Attr,pEuro))
												App_Data.Euro=(Val[0]!='0');
												else
													if (!strcmp(Attr,pExRom))
														App_Data.ExRom=(Val[0]!='0');
													else
														if (!strcmp(Attr,pBios))
															strcpy(Bios,Val);
														else
															if (!strcmp(Attr,pFps))
																Show_Fps=(Val[0]!='0');
															else
																if (!strcmp(Attr,pFullScr))
																	App_Data.FullScreen=(Val[0]!='0');
																else
																{
																	fprintf(stderr,"Invalid option : %s\n",Attr);
																	return 0;
																}
														
	return 1;
}

void Read_Default_Config(void)
{
	FILE *f;
	static char attr[MAXC], val[MAXC], s[MAXC];
	char *p, *fn;
	int i,l;
		
	fn="loem_def.cfg";
	f=fopen(fn,"r");
	
	if (!f)
	{
		fn = "LOEM_DEF.CFG";
		f=fopen(fn,"r");
	}
	
	if (!f)
		return;
	
	l=0;
	
	while (fgets(s,MAXC-1,f))
	{
		l++;
		p=s;
		
		while (*p && (isspace(*p)))
			p++;
		
		if (*p && (*p != '#'))
		{
			i=0;
			
			while (*p && (!isspace(*p)) && (*p != '='))
				attr[i++]=*p++;
			
			attr[i]=0;
			
			while (*p && (isspace(*p)))
				p++;
			
			i=0;
			
			if (*p == '=')
			{
				p++;
				
				while (*p && (isspace(*p)))
					p++;
				
				if (*p == '"')
				{
					p++;
					
					while (*p && (*p != '"') && (*p != '\n') && (*p != '\r'))
						val[i++]=*p++;
				}
				else
					while (*p && (!isspace(*p)))
						val[i++]=*p++;
			}
			
			val[i]=0;
			
			if (strlen(attr) > 0)
			{
				strlwr(attr);
				
				if (!Parse_Option(attr,val))
					printf("Error in the %s file at line number %d !\n\n",fn,l);
			}
		}
	}
	
	fclose(f);
}

static long FileSize(FILE *S)
{
	long cp, length;
	
	cp=ftell(S);
   	fseek(S, 0L, SEEK_END);
   	length=ftell(S);
   	fseek(S,cp,SEEK_SET);
	
   	return length;
}

static void Load_Bios(const char *BiosName)
{
	FILE *fn=NULL;
	static char s[MAXC+10];
	unsigned long crc;
	
	if ((!BiosName) || (strlen(BiosName) == 0))
	{
		strcpy(s,"o2rom.bin");
		fn=fopen("o2rom.bin","rb");
		
		if (!fn)
			fn=fopen("O2ROM.BIN","rb");		
	}
	else
		if ((BiosName[strlen(BiosName)-1]=='/') || (BiosName[strlen(BiosName)-1]=='\\') || 
			(BiosName[strlen(BiosName)-1]==':'))
		{
			strcpy(s,BiosName);
			strcat(s,"O2ROM.BIN");
			fn=fopen(s,"rb");
			
			if (!fn)
			{
				strcpy(s,BiosName);
				strcat(s,"o2rom.bin");
				fn = fopen(s,"rb");
			}
			else
			{
				strcpy(s,BiosName);
				fn=fopen(BiosName,"rb");
			}
		}
		
		if (!fn)
		{
			fprintf(stderr,"Error loading bios ROM (%s)\n",s);
			exit(EXIT_FAILURE);
		}
			
		if (fread(&Rom1[0],1024,1,fn) != 1)
		{
			fprintf(stderr,"Error loading bios ROM o2rom.bin\n");
			exit(EXIT_FAILURE);
		}
			
		fclose(fn);
			
		memcpy(&Rom2[0],&Rom1[0],1024);
		memcpy(&Rom3[0],&Rom1[0],1024);
		memcpy(&Rom4[0],&Rom1[0],1024);
			
		crc=Crc32_Buf(Rom1,1024);
			
		if (crc == 0x8016A315)
			printf("Odyssey2 bios ROM loaded\n");
		else
			if (crc == 0xE20A9F41)
				printf("Videopac+ G7400 bios ROM loaded\n");
			else
				printf("Bios ROM loaded (unknown version)\n");	
}

static void Load_Cart(char *File)
{
	FILE *fn=NULL;
	long l;
	
	App_Data.Crc=Crc32_File(File);
	
	if (App_Data.Crc == 0xAFB23F89)
		App_Data.ExRom=1;
	
	if (App_Data.Crc == 0x3BFEF56B)
		App_Data.ExRom=1;
	
	if (((App_Data.Crc == 0x975AB8DA) || (App_Data.Crc == 0xE246A812)) && (!App_Data.Debug))
	{
		fprintf(stderr,"Error: file %s is an incomplete ROM dump\n",File);
		exit(EXIT_FAILURE);
	}
	
	fn=fopen(File,"rb");
	
	if (!fn)
	{
		fprintf(stderr,"Error loading %s\n",File);
		exit(EXIT_FAILURE);
	}
	
	printf("Loading: \"%s\"  Size: ",File);
	l=FileSize(fn);
	
	if (l < 2049)
	{
		App_Data.Bank=1;
		
		if (fread(&Rom1[1024],l,1,fn) != 1)
		{
			fprintf(stderr,"Error loading %s\n",File);
			exit(EXIT_FAILURE);
		}
		
		memcpy(&Rom1[3072],&Rom1[2048],1024);
		Rom=Rom1;
		printf("2K");
	}
	else
		if (l == 3072)
		{
			App_Data.Bank=1;
			
			if (fread(&Rom1[1024],3072,1,fn) != 1)
			{
				fprintf(stderr,"Error loading %s\n",File);
				exit(EXIT_FAILURE);
			}
			
			Rom=Rom1;
			printf("3K");
		}
		else
			if (l == 4096)
			{
				if (App_Data.ExRom)
				{
					App_Data.Bank=1;
					
					if (fread(&ExtRom[0], 1024,1,fn) != 1)
					{
						fprintf(stderr,"Error loading %s\n",File);
						exit(EXIT_FAILURE);
					}
					
					if (fread(&Rom1[1024],3072,1,fn) != 1)
					{
						fprintf(stderr,"Error loading %s\n",File);
						exit(EXIT_FAILURE);
					}
					
					Rom=Rom1;
					printf("3K EXROM");
				}
				else
				{
					App_Data.Bank=2;
					
					if (fread(&Rom1[1024],2048,1,fn) != 1)
					{
						fprintf(stderr,"Error loading %s\n",File);
						exit(EXIT_FAILURE);
					}
					
					if (fread(&Rom2[1024],2048,1,fn) != 1)
					{
						fprintf(stderr,"Error loading %s\n",File);
						exit(EXIT_FAILURE);
					}
					
					memcpy(&Rom1[3072],&Rom1[2048],1024);
					memcpy(&Rom2[3072],&Rom2[2048],1024);
					Rom=Rom2;
					printf("4K");
				}
			}
			else
				if (l == 6144)
				{
					App_Data.Bank=2;
					
					if (fread(&Rom1[1024],3072,1,fn) != 1)
					{
						fprintf(stderr,"Error loading %s\n",File);
						exit(EXIT_FAILURE);
					}
					
					if (fread(&Rom2[1024],3072,1,fn) != 1)
					{
						fprintf(stderr,"Error loading %s\n",File);
						exit(EXIT_FAILURE);
					}
					
					Rom=Rom2;
					printf("6K");
				}
				else
					if (l == 8192)
					{
						App_Data.Bank=3;
						
						if (fread(&Rom1[1024],2048,1,fn) != 1)
						{
							fprintf(stderr,"Error loading %s\n",File);
							exit(EXIT_FAILURE);
						}
						
						if (fread(&Rom2[1024],2048,1,fn) != 1)
						{
							fprintf(stderr,"Error loading %s\n",File);
							exit(EXIT_FAILURE);
						}
						
						if (fread(&Rom3[1024],2048,1,fn) != 1)
						{
							fprintf(stderr,"Error loading %s\n",File);
							exit(EXIT_FAILURE);
						}
						
						if (fread(&Rom4[1024],2048,1,fn) != 1)
						{
							fprintf(stderr,"Error loading %s\n",File);
							exit(EXIT_FAILURE);
						}
						
						memcpy(&Rom1[3072],&Rom1[2048],1024);
						memcpy(&Rom2[3072],&Rom2[2048],1024);
						memcpy(&Rom3[3072],&Rom3[2048],1024);
						memcpy(&Rom4[3072],&Rom4[2048],1024);
						
						Rom=Rom4;
						printf("8K");
					}
					else
						if (l == 12288)
						{
							App_Data.Bank=3;
							
							if (fread(&Rom1[1024],3072,1,fn) != 1)
							{
								fprintf(stderr,"Error loading %s\n",File);
								exit(EXIT_FAILURE);
							}
							
							if (fread(&Rom2[1024],3072,1,fn) != 1)
							{
								fprintf(stderr,"Error loading %s\n",File);
								exit(EXIT_FAILURE);
							}
							
							if (fread(&Rom3[1024],3072,1,fn) != 1)
							{
								fprintf(stderr,"Error loading %s\n",File);
								exit(EXIT_FAILURE);
							}
							
							if (fread(&Rom4[1024],3072,1,fn) != 1)
							{
								fprintf(stderr,"Error loading %s\n",File);
								exit(EXIT_FAILURE);
							}
							
							Rom=Rom4;
							printf("12K");
						}
						else
						{
							fclose(fn);
							fprintf(stderr,"Invalid ROM size\n");
							exit(EXIT_FAILURE);
						}
						
	fclose(fn);
						
	if ((Rom1[1024+12]=='O') && (Rom1[1024+13]=='P') && 
		(Rom1[1024+14]=='N') && (Rom1[1024+15]=='B'))
		App_Data.OpenB=1;
	
	oRom=Rom;
	
	printf("  CRC: %08lX\n",App_Data.Crc);
}
