#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <allegro.h>

#include "tipi.h"
#include "vmachine.h"
#include "keyboard.h"
#include "cset.h"
#include "timefunc.h"
#include "cpu.h"
#include "vpp.h"
#include "vdc.h"

#define COL_SP0   		0x01
#define COL_SP1   		0x02
#define COL_SP2   		0x04
#define COL_SP3   		0x08
#define COL_VGRID 	0x10
#define COL_HGRID 	0x20
#define COL_VPP   		0x40
#define COL_CHAR  	0x80

#define X_START     	8
#define Y_START		24

static long ColorTable[16]={
	0x000000,
	0x0e3dd4,
	0x00981b,
	0x00bbd9,
	0xc70008,
	0xcc16b3,
	0x9d8710,
	0xe1dee1,
	0x5f6e6b,
	0x6aa1ff,
	0x3df07a,
	0x31ffff,
	0xff4255,
	0xff98ff,
	0xd9ad5d,
	0xffffff
};

static Byte *Col=NULL;
static PALETTE Colors,OldCol;
static Byte *VScreen=NULL;
static BITMAP *Bmp,*BmpCache;
static int Cached_Lines[MAXLINES];

Byte ColTab[256];
long Clip_Low;
long Clip_High;
int Show_Fps=0;


static void Create_CMap(void);
static void Draw_Char(Byte YPos,Byte XPos,Byte Chr,Byte Col);
static void Draw_Grid(void);
INLINE void MputVid(unsigned int Ad,unsigned int Len,Byte D,Byte C);


void Draw_Region(void)
{
	int I;
	
	if (RegionOff == 0xffff)
		I=Master_Clk / (LINECNT - 1) - 5;
	else
		I=Master_Clk / 22 + RegionOff;
	
	I=SnapLine(I,VDCWrite[0xA0],0);
	
	if (I < 0)
		I=0;
	
	Clip_Low=Last_Line * (long)BMPW;
	Clip_High=I * (long)BMPW;
	
	if (Clip_High > BMPW * BMPH)
		Clip_High=BMPW * BMPH;
	
	if (Clip_Low < 0)
		Clip_Low=0;
	
	if (Clip_Low < Clip_High)
		Draw_Display();
	
	Last_Line=I;
}

static void Create_CMap(void)
{
	register int i;
	
	for (i=0; i < 16; i++)
	{
		Colors[i+32].r=Colors[i].r=(ColorTable[i] & 0xff0000) >> 18;
		Colors[i+32].g=Colors[i].g=(ColorTable[i] & 0x00ff00) >> 10;
		Colors[i+32].b=Colors[i].b=(ColorTable[i] & 0x0000ff) >> 2;
	}
	
	for (i=16; i < 32; i++)
	{
		Colors[i+32].r=Colors[i].r=Colors[i-16].r / 2;
		Colors[i+32].g=Colors[i].g=Colors[i-16].g / 2;
		Colors[i+32].b=Colors[i].b=Colors[i-16].b / 2;
	}
	
	for (i=64; i < 256; i++)
		Colors[i].r=Colors[i].g=Colors[i].b=0;
}

void GrMode(void)
{
	set_color_depth(8);
	
	if (App_Data.FullScreen)
	{		
		if (App_Data.ScanLines)
		{
			App_Data.WSize=2;
			
			if (set_gfx_mode(GFX_AUTODETECT_FULLSCREEN,640,480,0,0))
			{
				App_Data.WSize=1;
				
				if (set_gfx_mode(GFX_AUTODETECT_FULLSCREEN,320,240,0,0))
				{
					fprintf(stderr,"Error: could not create screen.\n");
					exit(EXIT_FAILURE);
				}				
			}			
		}
		else
		{
			App_Data.WSize=2;
			
			if (set_gfx_mode(GFX_AUTODETECT_FULLSCREEN,640,480,0,0))
			{
				App_Data.WSize=1;
				
				if (set_gfx_mode(GFX_AUTODETECT_FULLSCREEN,320,240,0,0))
				{
					fprintf(stderr,"Error: could not create screen.\n");
					exit(EXIT_FAILURE);
				}
			}
		}
	}
	else
	{		
		if (set_gfx_mode(GFX_AUTODETECT_WINDOWED, WNDW * App_Data.WSize,WNDH * App_Data.WSize, 0, 0))
		{
			App_Data.WSize=2;
			
			if (set_gfx_mode(GFX_AUTODETECT_WINDOWED, WNDW*2, WNDH*2, 0, 0))
			{
				if (set_gfx_mode(GFX_AUTODETECT, WNDW*2, WNDH*2, 0, 0))
				{
					fprintf(stderr,"Error: could not create window.\n");
					exit(EXIT_FAILURE);
				}
			}
		}
	}
	
	set_palette(Colors);
	set_window_title(App_Data.Window_Title);
	ClearScr();
	set_display_switch_mode(SWITCH_PAUSE);
}

void Set_TextMode(void)
{
	set_palette(OldCol);
	set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
	
	if (New_Int)
		Set_Old_Int9();
}

void ClearScr(void)
{
	acquire_screen();
	clear(screen);
	release_screen();
	clear(BmpCache);
}

INLINE void MputVid(unsigned int Ad,unsigned int Len,Byte D,Byte C)
{
	if ((Ad > (unsigned long)Clip_Low) && (Ad < (unsigned long)Clip_High))
	{
		unsigned int i;
		
		if (((Len & 3) == 0) && (sizeof(unsigned long) == 4))
		{
			unsigned long dddd=(((unsigned long)D) & 0xff) | ((((unsigned long)D) & 0xff) << 8) | 
												((((unsigned long)D) & 0xff) << 16) | ((((unsigned long)D) & 0xff) << 24);
			unsigned long cccc=(((unsigned long)C) & 0xff) | ((((unsigned long)C) & 0xff) << 8) | 
												((((unsigned long)C) & 0xff) << 16) | ((((unsigned long)C) & 0xff) << 24);
			
			for (i=0;  i < Len >> 2; i++)
			{
				*((unsigned long*)(VScreen + Ad))=dddd;
				cccc|=*((unsigned long*)(Col + Ad));
				*((unsigned long*)(Col + Ad))=cccc;
				ColTab[C]|=((cccc | (cccc >> 8) | (cccc >> 16) | (cccc >> 24)) & 0xff);
				Ad+=4;
			}
		}
		else
			for (i=0; i < Len; i++)
			{
				VScreen[Ad]=D;
				Col[Ad]|=C;
				ColTab[C]|=Col[Ad++];
			}
	}
}

static void Draw_Grid(void)
{
	unsigned int Pnt,Pn1;
	Byte mask,D;
	int j,i,x,w;
	Byte Color;
	
	if (VDCWrite[0xA0] & 0x40)
		for(j=0; j < 9; j++)
		{
			Pnt=(((j * 24) + 24) * BMPW);
			
			for (i=0; i < 9; i++)
			{
				Pn1=Pnt + (i * 32) + 20;
				Color=ColorVector[j * 24 + 24];
				
				MputVid(Pn1, 4, (Color & 0x07) | ((Color & 0x40) >> 3) | (Color & 0x80 ? 0 : 8),COL_HGRID);
				
				Color=ColorVector[j * 24 + 25];
				
				MputVid(Pn1 + BMPW, 4,(Color & 0x07) | ((Color & 0x40) >> 3) | (Color & 0x80 ? 0 : 8),COL_HGRID);
				
				Color=ColorVector[j * 24 + 26];
				
				MputVid(Pn1 + BMPW * 2, 4, (Color & 0x07) | ((Color & 0x40) >> 3) | (Color & 0x80 ? 0 : 8),COL_HGRID);
			}
		}
		
	mask=0x01;
		
	for(j=0; j < 9; j++)
	{
		Pnt=((( j *24) + 24) * BMPW);
		
		for (i=0; i < 9; i++)
		{
			Pn1=Pnt + (i * 32) + 20;
			
			if ((Pn1 + BMPW * 3 >= (unsigned long)Clip_Low) && (Pn1 <= (unsigned long)Clip_High))
			{
				D=VDCWrite[0xC0 + i];
				
				if (j == 8)
				{
					D=VDCWrite[0xD0 + i];
					mask=1;
				}
				
				if (D & mask)
				{
					Color=ColorVector[j * 24 + 24];
					
					MputVid(Pn1,36,(Color & 0x07) | ((Color & 0x40) >> 3) | (Color & 0x80 ? 0 : 8),COL_HGRID);
					
					Color=ColorVector[j * 24 + 25];
					
					MputVid(Pn1 + BMPW,36,(Color & 0x07) | ((Color & 0x40) >> 3) | (Color & 0x80 ? 0 : 8),COL_HGRID);
					
					Color=ColorVector[j * 24 + 26];
					
					MputVid(Pn1 + BMPW * 2,36,(Color & 0x07) | ((Color & 0x40) >> 3) | (Color & 0x80 ? 0 : 8),COL_HGRID);
				}
			}
		}
		
		mask<<=1;
	}
	
	mask=0x01;
	w=4;
	
	if (VDCWrite[0xA0] & 0x80)
		w=32;
	
	for(j=0; j < 10; j++)
	{
		Pnt=(j * 32);
		mask=0x01;
		D=VDCWrite[0xE0 + j];
		
		for (x=0; x < 8; x++)
		{
			Pn1=Pnt + (((x * 24) + 24) * BMPW) + 20;
			
			if (D & mask)
			{
				for(i=0; i < 24; i++)
				{
					if ((Pn1 >= (unsigned long)Clip_Low) && (Pn1 <= (unsigned long)Clip_High))
					{
						Color=ColorVector[x * 24 + 24 + i];
						
						MputVid(Pn1,w,(Color & 0x07) | ((Color & 0x40) >> 3) | (Color & 0x80 ? 0 : 8),COL_VGRID);
					}
					
					Pn1+=BMPW;
				}
			}
			
			mask<<=1;
		}
	}
}

void Finish_Display()
{
	int x,y,sn;
	static int Cache_Counter=0;
	
	Vpp_Finish_Bmp(VScreen,9,5,BMPW - 9,BMPH - 5,Bmp->w,Bmp->h);
	
	if (Show_Fps)
	{
		static long last=-1, index=0, curr=0, t=0;
		
		if (last<0)
			last=GetTimeTicks();
		
		index=(index + 1) % 200;
		
		if (!index)
		{
			t=GetTimeTicks();
			curr=t - last;
			last=t;
		}
		
		if (curr)
		{
			text_mode(0);
			textprintf(Bmp,font, 20,4,7,"FPS: %3d",(int)((200.0*TICKSPERSEC)/curr+0.5));
		}
	}
	
	for (y=0; y < Bmp->h; y++)
	{
		Cached_Lines[y]=!memcmp(BmpCache->line[y],Bmp->line[y],Bmp->w);
		
		if (!Cached_Lines[y])
			memcpy(BmpCache->line[y],Bmp->line[y],Bmp->w);
	}
	
	for (y=0; y < 10; y++)
		Cached_Lines[(y + Cache_Counter) % Bmp->h]=0;
	
	Cache_Counter=(Cache_Counter + 10) % Bmp->h;
	
	acquire_screen();
	
	sn=((App_Data.WSize>1) && (App_Data.ScanLines)) ? 1 : 0;
	
	for (y=0; y < WNDH; y++)
		if (!Cached_Lines[y + 2])
			stretch_blit(Bmp,screen,7,2+y,WNDW,1,0,y * App_Data.WSize,WNDW * App_Data.WSize,
		                      App_Data.WSize - sn);
		
	if (sn)
		for (y=0; y < WNDH; y++)
			if (!Cached_Lines[y + 2])
			{
				for (x=0; x < Bmp->w; x++)
					Bmp->line[y+2][x]+=16;
				
				stretch_blit(Bmp,screen,7,2+y,WNDW,1,0,(y+1)*App_Data.WSize-1,WNDW*App_Data.WSize,1);
				memcpy(Bmp->line[y+2], BmpCache->line[y+2],Bmp->w);
			}
			
	release_screen();
}

void Clear_Collision(void)
{
	Load_ColPlus(Col);
	
	ColTab[0x01]=ColTab[0x02]=0;
	ColTab[0x04]=ColTab[0x08]=0;
	ColTab[0x10]=ColTab[0x20]=0;
	ColTab[0x40]=ColTab[0x80]=0;
}

void Draw_Display(void)
{
	int i,j,x,sm,t;
	Byte y,xt,yt,b,d1,cl,c;
	unsigned int Pnt,Pnt2;
	
	for (i=Clip_Low / BMPW; i < Clip_High / BMPW; i++)
		memset(VScreen + i * BMPW,((ColorVector[i] & 0x38) >> 3) | (ColorVector[i] & 0x80 ? 0 : 8), BMPW);
	
	if (VDCWrite[0xA0] & 0x08)
		Draw_Grid();
	
	if (UseForEn && (!(VDCWrite[0xA0] & 0x20)))
		return;
	
	for(i=0x10; i < 0x40; i+=4)
		Draw_Char(VDCWrite[i],VDCWrite[i+1],VDCWrite[i+2],VDCWrite[i+3]);
	
	Pnt=0x40;
	
	for(i=0; i < 4; i++)
	{
		x=y=248;
		
		for (j=0; j < 4; j++)
		{
			xt=VDCWrite[Pnt + j * 4 + 1];
			yt=VDCWrite[Pnt + j * 4];
			
			if ((xt < 240) && (yt < 240))
			{
				x=xt;
				y=yt;
				
				break;
			}
		}
		
		for(j=0; j < 4; j++)
		{
			Draw_Char(y,x,VDCWrite[Pnt + 2],VDCWrite[Pnt + 3]);
			x+=16;
			Pnt+=4;
		}
	}
	
	c=8;
	
	for (i=12; i >= 0; i -=4)
	{
		Pnt2=0x80 + (i * 2);
		y=VDCWrite[i];
		x=VDCWrite[i + 1] - 8;
		t=VDCWrite[i + 2];
		cl=((t & 0x38) >> 3);
		cl=((cl & 2) | ((cl & 1) << 2) | ((cl & 4)>> 2)) + 8;
		
		if ((x < 164) && (y > 0) && (y < 232))
		{
			Pnt=y * BMPW + (x * 2) + 20 + SprOff;
			
			if (t & 4)
			{
				if ((Pnt + BMPW * 32 >= (unsigned long)Clip_Low) && (Pnt <= (unsigned long)Clip_High))
				{
					for (j=0; j < 8; j++)
					{
						sm=(((j % 2 == 0) && (((t >> 1) & 1) != (t & 1))) || ((j%2==1) && (t & 1))) ? 1 : 0;
						d1=VDCWrite[Pnt2++];
						
						for (b=0; b < 8; b++)
						{
							if (d1 & 0x01)
							{
								if ((x + b + sm < 159) && (y + j < 247))
								{
									MputVid(sm + Pnt,4,cl,c);
									MputVid(sm + Pnt + BMPW,4,cl,c);
									MputVid(sm + Pnt + 2 * BMPW,4,cl,c);
									MputVid(sm + Pnt + 3 * BMPW,4,cl,c);
								}
							}
							
							Pnt+=4;
							d1>>=1;
						}
						
						Pnt+=BMPW * 4 - 32;
					}
				}
			}
			else
			{
				if ((Pnt + BMPW * 16 >= (unsigned long)Clip_Low) && (Pnt <= (unsigned long)Clip_High))
				{
					for (j=0; j < 8; j++)
					{
						sm=(((j % 2==0) && (((t >> 1) & 1) != (t & 1))) || ((j % 2==1) && (t & 1))) ? 1 : 0;
						d1=VDCWrite[Pnt2++];
						
						for (b=0; b < 8; b++)
						{
							if (d1 & 0x01)
							{
								if ((x + b + sm < 160) && (y + j < 249))
								{
									MputVid(sm + Pnt,2,cl,c);
									MputVid(sm + Pnt + BMPW,2,cl,c);
								}
							}
							
							Pnt+=2;
							d1>>=1;
						}
						
						Pnt+=BMPW * 2 - 16;
					}
				}
			}
		}
		
		c>>=1;
	}
}

static void Draw_Char(Byte YPos,Byte XPos,Byte Chr,Byte Col)
{
	int j,c;
	Byte cl,d1;
	int y,b,n;
	unsigned int Pnt;
	
	y=(YPos & 0xFE);
	Pnt=y * BMPW + ((XPos - 8) * 2) + 20;
	YPos>>=1;
	n=8 - (YPos % 8) - (Chr % 8);
	
	if (n < 3)
		n+=7;
	
	if ((Pnt + BMPW * 2 * n >= (unsigned long)Clip_Low) && (Pnt <= (unsigned long)Clip_High))
	{
		c=(int)Chr + YPos;
		
		if (Col & 0x01)
			c+=256;
		
		if (c > 511)
			c-=512;
		
		cl=((Col & 0x0E) >> 1);
		cl=((cl & 2) | ((cl & 1) << 2) | ((cl & 4) >> 2)) + 8;
		
		if ((y > 0) && (y < 232) && (XPos < 157))
		{
			for (j=0; j < n; j++)
			{
				d1=CSet[c + j];
				
				for (b=0; b < 8; b++)
				{
					if (d1 & 0x80)
					{
						if ((XPos - 8+ b < 160) && (y + j < 240))
						{
							MputVid(Pnt,2,cl,COL_CHAR);
							MputVid(Pnt + BMPW,2,cl,COL_CHAR);
						}
					}
					
					Pnt+=2;
					d1<<=1;
				}
				Pnt+=BMPW * 2 - 16;
			}
		}
	}
}

void Close_Display(void)
{
	free(VScreen);
	free(Col);
}

void Window_Close_Hook(void)
{
	Key_Debug=0;
	Key_Done=1;
	
	allegro_exit();
}

static void TxtMsg(int X,int Y,int C,const char *S)
{
	text_mode(-1);
	
	textout_centre(Bmp,font,S,X + 1 ,Y + 1,32);
	textout_centre(Bmp,font,S,X,Y,C);
}

void Display_Info(void)
{
	rectfill(Bmp,20,72,311,158,8+32);
	line(Bmp,20,72,311,72,15+32);
	line(Bmp,20,72,20,158,15+32);
	line(Bmp,21,158,311,158,1+32);
	line(Bmp,311,158,311,72,1+32);
	
	TxtMsg(166,76,15+32,"loem .:Linux Odyssy2 EMulator:.");
	TxtMsg(166,90,15+32,"Based on O2EM");	
	TxtMsg(166,118,15+32,"loem (c) Calogiuri Enzo Antonio");
	TxtMsg(166,132,15+32,"mail me at enzo76@freemail.it");
	TxtMsg(166,148,15+32,"Press F1 to continue");
	
	Finish_Display();
}

void Init_Display(void)
{
	get_palette(OldCol);
	
	Create_CMap();
	
	Bmp=create_bitmap(BMPW,BMPH);
	BmpCache=create_bitmap(BMPW,BMPH);
	
	if ((!Bmp) || (!BmpCache))
	{
		fprintf(stderr,"Could not allocate memory for screen buffer.\n");
		exit(EXIT_FAILURE);
	}
	
	VScreen=(Byte *) Bmp->dat;
	
	clear(Bmp);
	clear(BmpCache);
	
	Col=(Byte *)malloc(BMPW*BMPH);
	
	if (!Col)
	{
		fprintf(stderr,"Could not allocate memory for collision buffer.\n");
		free(VScreen);
		exit(EXIT_FAILURE);
	}
	
	memset(Col,0,BMPW*BMPH);
	
	if (!App_Data.Debug)
	{
		GrMode();
		
		Init_Keyboard();
	}
	
	set_window_close_button(TRUE);
	set_window_close_hook(Window_Close_Hook);
}
