#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <allegro.h>

#include "vmachine.h"
#include "vdc.h"
#include "vpp_cset.h"
#include "vpp.h"

static void Vpp_Draw_Char(int X,int Y,Byte Ch,Byte C0,Byte C1,
                                               Byte Ext,Byte Dw,Byte Dh,Byte Ul);
static void Vpp_UpDate_Screen(void);

static Byte LumReg=0xff,TraReg=0xff;
static BITMAP *VppBmp=NULL;
static Byte *ColPlus=NULL;
static int VppOn=1;
static int Vpp_Cx=0;
static int Vpp_Cy=0;
static Byte Vpp_Data=0;
static int Inc_Curs=1;
static int Slice=0;
static int Vpp_Y0=0;
static Byte Vpp_R=0;
static Byte dChars[2][960];
static Byte Vpp_Mem[40][32][4];
static int Frame_Cnt=0;
static int Blink_St=0;
static int SliceMode=0;
static int Need_UpDate=0;


Byte Read_PB(Byte P)
{
	P&=0x3;
	
	switch (P)
	{
		case 0:
			return LumReg >> 4;
		
		case 1:
			return LumReg & 0xF;
		
		case 2:
			return TraReg >> 4;
		
		case 3:
			return TraReg & 0xF;
	}
	
	return 0;
}

void Write_PB(Byte P,Byte Val)
{
	P&=0x3;
	Val&=0xF;
	
	switch (P)
	{
		case 0:
			LumReg=(Val << 4) | (LumReg & 0xF);
			break;
		
		case 1:
			LumReg=(LumReg & 0xF0) | Val;
			break;
		
		case 2:
			TraReg=(Val << 4) | (TraReg & 0xF);
			break;
		
		case 3:
			TraReg=(TraReg & 0xF0) | Val;
			break;
	}
	
	Need_UpDate=1;
}

Byte Vpp_Read(Indirizzo Adr)
{
	register Byte t;
	
	switch (Adr)
	{
		case 4:
			return Vpp_Mem[Vpp_Cx][Vpp_Cy][1];
		
		case 5:
			if (SliceMode)
			{
				Byte ext,chr;
				
				chr=Vpp_Mem[Vpp_Cx][Vpp_Cy][0];
				ext=(Vpp_Mem[Vpp_Cx][Vpp_Cy][1] & 0x80) ? 1 : 0;
				
				if (chr < 0xA0)
					t=0;
				else
				{
					t=dChars[ext][(chr - 0xA0) * 10 +  Slice];
					t =((t & 0x80) >> 7) | ((t & 0x40) >> 5) | ((t & 0x20) >> 3) | 
					     ((t & 0x10) >> 1) | ((t & 0x08) << 1) | ((t & 0x04) << 3) | 
					     ((t & 0x02) << 5) | ((t & 0x01) << 7);
				}
				
				Slice=(Slice  + 1) % 10;
			}
			else
			{
				t=Vpp_Mem[Vpp_Cx][Vpp_Cy][0];
				
				if (Inc_Curs)
				{
					Vpp_Cx++;
					
					if (Vpp_Cx >= 40)
					{
						Vpp_Cx=0;
						Vpp_Cy++;
						
						if (Vpp_Cy >= 24)
							Vpp_Cy=0;
					}
				}
			}
			return t;
			
		case 6:
			return 0;
		
		default:
			return 0;
	}
}

void Vpp_Write(Byte Dat,Indirizzo Adr)
{
	switch (Adr)
	{
		case 0:
			if (!SliceMode)
				Vpp_Mem[Vpp_Cx][Vpp_Cy][1]=Dat;
			
			break;
			
		case 1:
			if (SliceMode)
			{
				Byte ext,chr;
				
				chr=Vpp_Mem[Vpp_Cx][Vpp_Cy][0];
				ext=(Vpp_Mem[Vpp_Cx][Vpp_Cy][1] & 0x80) ? 1 : 0;
				
				if (chr >= 0xA0)
					dChars[ext][(chr - 0xA0) * 10 + Slice]=((Dat & 0x80) >> 7) | ((Dat & 0x40) >> 5) | 
				    ((Dat & 0x20) >> 3) | ((Dat & 0x10) >> 1) | ((Dat & 0x08) << 1) | ((Dat & 0x04) << 3) | 
				    ((Dat & 0x02) << 5) | ((Dat & 0x01) << 7);
				
				Slice=(Slice+1) % 10;
			}
			else
			{
				Vpp_Mem[Vpp_Cx][Vpp_Cy][0]=Dat;
				
				if ((Dat > 0x7f) && (Dat < 0xa0) && (!(Vpp_Mem[Vpp_Cx][Vpp_Cy][1] & 0x80)))
				{
					Vpp_Mem[Vpp_Cx][Vpp_Cy][2]=Dat;
					Vpp_Mem[Vpp_Cx][Vpp_Cy][3]=Vpp_Mem[Vpp_Cx][Vpp_Cy][1];
				}
				else
					Vpp_Mem[Vpp_Cx][Vpp_Cy][2]=Vpp_Mem[Vpp_Cx][Vpp_Cy][3]=0;
				
				if (Inc_Curs)
				{
					Vpp_Cx++;
					
					if (Vpp_Cx >= 40)
					{
						Vpp_Cx=0;
						Vpp_Cy++;
						
						if (Vpp_Cy >= 24)
							Vpp_Cy=0;
					}
				}
			}
			break;
			
		case 2:
			Vpp_Data=Dat;
			break;
		
		case 3:
			switch (Dat & 0xE0)
			{
				case 0x00:
					Vpp_Cy=Vpp_Data & 0x1F;
					Vpp_Cx=0;
					break;
				
				case 0x20:
					Vpp_Cy=Vpp_Data & 0x1f;
					break;
				
				case 0x40:
					Vpp_Cx=Vpp_Data % 40;
					break;
				
				case 0x60:
					Vpp_Cx++;
				
					if (Vpp_Cx >= 40)
					{
						Vpp_Cx=0;
						Vpp_Cy++;
						
						if (Vpp_Cy >= 24)
							Vpp_Cy=0;
					}
					
					break;
					
				case 0x80:
					SliceMode=0;
					Slice=(Vpp_Data & 0x1F) % 10;
				
					switch (Vpp_Data & 0xE0)
					{
						case 0x00:
							Inc_Curs = 1;
							break;
						
						case 0x20:
							Inc_Curs = 1;
							break;
						
						case 0x40:
							Inc_Curs = 0;
							break;
						
						case 0x60:
							Inc_Curs = 0;
							break;
						
						case 0x80:
							SliceMode=1;
							break;
						
						case 0xA0:
							SliceMode=1;
							break;
						
						default:
							break;
					}
					break;
					
				case 0xA0:
					Vpp_R=Vpp_Data;
					break;
				
				case 0xC0:
					Vpp_Y0=(Vpp_Data & 0x1f) % 24;
					break;
				
				default:
					break;
			}
			break;
			
		default:
			break;
	}
	
	Need_UpDate=1;
}

void Vpp_Finish_Bmp(Byte *VMem,int Offx,int Offy,int W,int H,int Totw,int Toth)
{
	int i,x,y,t,c,nc,clrx,clry;
	int tcol[16],m[8]={0x01, 0x10, 0x04, 0x40, 0x02, 0x20, 0x08, 0x80};
	Byte *pnt,*pnt2,*pnt3;
	
	if (VppOn)
	{
		memset(ColPlus,0,BMPW*BMPH);
		VppOn=0;
	}
	
	if (TraReg == 0xFF)
		return;
	
	VppOn=1;
	Frame_Cnt--;
	
	if (Frame_Cnt <= 0)
	{
		Frame_Cnt=100;
		Blink_St=1 - Blink_St;
		Need_UpDate=1;
	}
	
	if (Need_UpDate)
		Vpp_UpDate_Screen();
	
	for (i=0; i < 8; i++)
		tcol[i]=tcol[i+8]=!(TraReg & m[i]);
	
	if (W > Totw - Offx)
		W=Totw - Offx;
	
	if (H > Toth - Offy)
		H=Toth - Offy;
	
	if (W > VppBmp->w)
		W=VppBmp->w;
	
	if (H > VppBmp->h)
		H=VppBmp->h;
	
	clrx=clry=0;
	
	for (i=0; (!clrx) && (i < Totw); i++)
		if (tcol[VMem[Offy * Totw + i] & 7])
			clrx=1;
		
	for (i=0; (!clry) && (i < Toth); i++)
		if (tcol[VMem[i * Totw + Offx] & 7])
			clry=1;
		
	if (clrx)
		for (y=0; y < Offy; y++)
			for (x=0; x < Totw; x++)
				VMem[y * Totw + x]=0;
			
	if (clry)
		for (y=0; y < Toth; y++)
			for (x=0; x < Offx; x++)
				VMem[y * Totw + x]=0;
			
	for (y=0; y < H; y++)
	{
		pnt=VMem + (Offy + y) * Totw + Offx;
		pnt2=(Byte *)VppBmp->line[y];
		x=0;
		
		while (x < W)
		{
			pnt3=pnt;
			c=*pnt++;
			t=x++;
			
			if ((((x + Offx) & 3) == 0) && (sizeof(unsigned long) == 4))
			{
				unsigned long cccc, dddd,*p = (unsigned long*)pnt;
				int t2=x, w2=W-4;
				
				cccc=(((unsigned long)c) & 0xff) | ((((unsigned long)c) & 0xff) << 8) | 
				          ((((unsigned long)c) & 0xff) << 16) | ((((unsigned long)c) & 0xff) << 24);
				dddd=*p++;
				
				while ((x < w2) && (dddd == cccc))
				{
					x+=4;
					dddd=*p++;
				}
				
				pnt+=x - t2;
			}
			
			if (c < 16)
			{
				if (tcol[c])
				{
					if (App_Data.OpenB)
						for (i=0; i < x - t; i++)
							*pnt3++=*pnt2++ & 0xf;
					else
					{
						memcpy(pnt3,pnt2,x-t);
						pnt2+=x - t;
					}
				}
				else
				{
					for (i=0; i < x - t; i++)
					{
						nc=*pnt2++;
						
						if ((nc & 0x10) && App_Data.OpenB)
							*pnt3++=nc & 0xf;
						else
							if (nc & 8)
								ColPlus[pnt3++ - VMem]=0x40;
							else
								pnt3++;
					}
				}
			}
		}
	}
}

static void Vpp_Draw_Char(int X,int Y,Byte Ch,Byte C0,Byte C1,
                                               Byte Ext,Byte Dw,Byte Dh,Byte Ul)
{
	int xx,yy,d,m,k;
	
	if ((X > 39) || (Y > 24) || (Ext > 1))
		return;
	
	d=(Dh == 2) ? 5 : 0;
	
	for (yy=0; yy < 10; yy++)
	{
		if (Ul && (d == 9))
			k=255;
		else
			if (Ch >= 0xA0)
				k=dChars[Ext][(Ch - 0xA0) * 10 + d];
			else
				if (Ch >= 0x80)
					k=255;
				else
					k=Vpp_CSet[Ext][Ch * 10 + d];
				
		m=(Dw == 2) ? 0x08 : 0x80;
				
		for (xx=0; xx < 8; xx++)
		{
			VppBmp->line[Y * 10 + yy][X * 8 + xx]=(k & m) ? C1 : C0;
			
			if ((xx % 2) || (Dw == 0))
				m>>=1;
		}
		
		if ((yy % 2) || (Dh == 0))
			d++;
	}
}

static void Vpp_UpDate_Screen(void)
{
	int i,x,y,l,chr,attr,ext,c0,c1,dw,dh,hpar,vpar,lvd,lhd,ser_chr,ser_atr,ul,conc,box;
	int tlum[8], m[8] = {0x01, 0x10, 0x04, 0x40, 0x02, 0x20, 0x08, 0x80};
	
	clear(VppBmp);
	
	for (i=0; i < 8; i++)
		tlum[i]=(LumReg & m[i]) ? 0 : 8;
	
	vpar=lvd=0;
	
	for (y=0; y < 25; y++)
	{
		vpar=(lvd == 0) ? 0 : 1 - vpar;
		l =(y == 0) ? 31 : (y - 1 + Vpp_Y0) % 24;
		c0=ul=conc=box=0;
		hpar=lhd=0;
		
		for (x=0; x < 40; x++)
		{
			hpar=(lhd == 0) ? 0 : 1 - hpar;
			
			chr=Vpp_Mem[x][l][0];
			attr=Vpp_Mem[x][l][1];
			c1=attr & 0x7;
			c1=((c1 & 2) | ((c1 & 1) << 2) | ((c1 & 4) >> 2));
			ext=(attr & 0x80) ? 1 : 0;
			ser_chr=Vpp_Mem[x][l][2];
			ser_atr=Vpp_Mem[x][l][3];
			
			if (ser_chr)
			{
				c0=(ser_atr >> 4) & 0x7;
				c0=((c0 & 2) | ((c0 & 1) << 2) | ((c0 & 4) >> 2));
				ul=ser_chr & 4;
				conc=ser_chr & 1;
				box=ser_chr & 2;
			}
			
			if (ext)
			{
				c0=(attr >> 4) & 0x7;
				c0=((c0 & 2) | ((c0 & 1) << 2) | ((c0 & 4) >> 2));
				dw=dh=0;
			}
			else
			{
				dw=(attr & 0x20) ? (hpar ? 2 : 1) : 0;
				dh=(attr & 0x10) ? (vpar ? 2 : 1) : 0;
				
				if (dw)
					lhd=1;
				
				if (dh)
					lvd=1;
			}
			
			if ((Vpp_R & 0x80) && (!(attr & 8)) && (!Blink_St))
				c1=c0;
			
			if (((y == 0) && (Vpp_R & 8)) || ((y != 0) && (Vpp_R & 1)))
			{
				if ((!conc) || (!(Vpp_R & 4)))
				{
					if (box || (!(Vpp_R & 2)))
					{
						if ((!ext) && (attr & 0x40))
							Vpp_Draw_Char(x,y,chr,c1 | tlum[c1],c0 | tlum[c0],ext,dw,dh,ul);
						else
							Vpp_Draw_Char(x,y,chr,c0 | tlum[c0],c1 | tlum[c1],ext,dw,dh,ul);
					}
					else
						Vpp_Draw_Char(x,y,255,(App_Data.OpenB) ? 16 : 0, 0, 0, 0, 0, 0);
				}
			}
		}
	}
	
	if (Vpp_R & 0x20)
		for (y = VppBmp->h-1; y >= 10; y--)
			for (x = 0; x < VppBmp->w; x++)
				VppBmp->line[y][x]=VppBmp->line[(y - 10) / 2 + 10][x];
			
	Need_UpDate=0;
}

void Load_ColPlus(Byte *Col)
{
	if (VppOn)
		memcpy(Col,ColPlus,BMPW*BMPH);
	else
		memset(Col,0,BMPW*BMPH);
}

void Init_Vpp(void)
{
	int i,j,k;
	
	if (!VppBmp)
		VppBmp=create_bitmap(320,250);
	
	if (!ColPlus)
		ColPlus=(Byte *)malloc(BMPW*BMPH);
	
	if ((!VppBmp) || (!ColPlus))
	{
		fprintf(stderr,"Could not allocate memory for Videopac+ screen buffer.\n");
		exit(EXIT_FAILURE);
	}
	
	clear(VppBmp);
	memset(ColPlus,0,BMPW*BMPH);
	
	LumReg=TraReg=0xFF;
	Vpp_Cx=0;
	Vpp_Cy=0;
	Vpp_Y0=0;
	Vpp_R=0;
	Inc_Curs=1;
	Vpp_Data=0;
	Frame_Cnt=0;
	Blink_St=0;
	Slice=0;
	SliceMode=0;
	Need_UpDate=1;
	VppOn=1;
	
	for (i=0; i < 2; i++)
		for (j=0; j < 960; j++)
			dChars[i][j]=0;
		
	for (i=0; i < 40; i++)
		for (j=0; j < 32; j++)
			for (k=0; k < 4; k++)
				Vpp_Mem[i][j][k]=0;
}
